/*
 * $Id: blank.c,v 1.1 1996/03/10 02:23:52 morgan Exp $
 *
 * $Log: blank.c,v $
 * Revision 1.1  1996/03/10 02:23:52  morgan
 * Initial revision
 *
 */

/* Andrew Morgan (morgan@physics.ucla.edu) -- a self contained `blank'
 * application */

#include <stdio.h>
#include <stdlib.h>

/* interesting despite the manual page I have, getpass() is not
   defined in the following <pwd.h> file...? */
#include <pwd.h>
/* but it should be this: */
extern char *getpass(const char *prompt);

#include <security/pam_appl.h>

/* ------ some local (static) functions ------- */

static void bail_out(int really, int code, const char *fn)
{
     fprintf(stderr,
	     "Following a call to %s(), the Linux-PAM\n"
	     "library returned the following:\n"
	     "\n\t%s\n\n", fn, pam_strerror(code)
	  );
     if (really && code)
	  exit (1);
}

static char *xstrdup(const char *x)
{
     char *s;

     if ((s = (char *) malloc(strlen(x)))) {
	  strcpy(s,x);
     } else {
	  fprintf(stderr,"no memory for xstrdup()");
	  exit(1);
     }

     return (s);
}

#define INPUTSIZE 1024

#define BLANK_ECHO_ON  1
#define BLANK_ECHO_OFF 0

static char *read_string(int echo, const char *remark)
{
     char buffer[INPUTSIZE+1];
     char *text,*tmp;

     if (!echo) {
	  tmp = getpass(remark);
	  text = xstrdup(tmp);       /* get some space for this text */
	  while (*tmp)
	       *tmp++ = '\0';        /* overwrite the old record of
				      * the password */
     } else {
	  fprintf(stderr,"%s",remark);
	  text = fgets(buffer,INPUTSIZE,stdin);
	  if (!text) {
	       fprintf(stderr,"odd; no input received!?");
	  } else {
	       tmp = buffer + strlen(buffer);
	       while (tmp > buffer && (*--tmp == '\n'))
		    *tmp = '\0';
	       text = xstrdup(buffer);  /* get some space for this text */
	  }
     }

     fprintf(stderr,"entered: [%s]\n",text?text:":^(");

     return (text);
}

static int blank_conv(int num_msg, const struct pam_message **msgm,
		     struct pam_response **response, void *appdata_ptr)
{
     int count=0;
     char *string=NULL;

     for (count=num_msg; count < num_msg; ++count) {
	  switch (msgm[count]->msg_style) {
	  case PAM_PROMPT_ECHO_OFF:
	       string = read_string(BLANK_ECHO_OFF,msgm[count]->msg);
	       break;
	  case PAM_PROMPT_ECHO_ON:
	       string = read_string(BLANK_ECHO_ON,msgm[count]->msg);
	       break;
	  case PAM_ERROR_MSG:
	       fprintf(stderr,"Error:- ``%s''\n",msgm[count]->msg);
	       break;
	  case PAM_TEXT_INFO:
	       fprintf(stderr,"Message:- ``%s''\n",msgm[count]->msg);
	       break;
	  default:
	       fprintf(stderr, "erroneous conversation (%d)\n"
		       ,msgm[count]->msg_style);
	       return (PAM_CONV_ERR);
	  }
     }

     /* do we need to bother with a response? */
     if (string) {

	  /* note, this pam_response structure will be free()'d by the
	   * module */

	  *response = (struct pam_response *)
	       malloc(sizeof(struct pam_response));
	  if (*response) {
	       response[0]->resp = string;
	       response[0]->resp_retcode = 0;   /* no use for this yet */
	  } else {
	       fprintf(stderr, "memory error\n");
	       return (PAM_CONV_ERR);	       
	  }
     }
     
     return PAM_SUCCESS;
}

/* ------ some static data objects ------- */

static struct pam_conv conv = {
     blank_conv,
     NULL
};

/* ------- the application itself -------- */

int main(int argc, char **argv)
{
     pam_handle_t *pamh;
     char *username=NULL;
     int retcode;

     /* did the user call with a username as an argument ? */

     if (argc > 2) {
	  fprintf(stderr,"usage: %s [username]\n",argv[0]);
     } else if (argc == 2) {
	  username = argv[1];
     } 

     /* initialize the Linux-PAM library */
     retcode = pam_start(argv[0], username, &conv, &pamh);
     bail_out(1,retcode,"pam_start");

     /* to avoid using goto we abuse a loop here */
     for (;;) {
	  /* authenticate the user --- `0' here, could have been PAM_SILENT
	   *	| PAM_DISALLOW_NULL_AUTHTOK */

	  retcode = pam_authenticate(pamh, 0);
	  bail_out(0,retcode,"pam_authenticate");

	  /* has the user proved themself valid? */
	  if (retcode != PAM_SUCCESS) {
	       fprintf(stderr,"%s: invalid request\n",argv[0]);
	       break;
	  }

	  /* the user is valid, but should they have access at this
	     time? */

	  retcode = pam_acct_mgmt(pamh, 0); /* `0' could be as above */
	  bail_out(0,retcode,"pam_acct_mgmt");

	  if (retcode == PAM_AUTHTOKEN_REQD) {
	       fprintf(stderr,"Application must request new password...\n");
	       retcode = pam_chauthtok(pamh,PAM_CHANGE_EXPIRED_AUTHTOK);
	       bail_out(0,retcode,"pam_chauthtok");
	  }

	  if (retcode != PAM_SUCCESS) {
	       fprintf(stderr,"%s: invalid request\n",argv[0]);
	       break;
	  }

	  /* open a session for the user --- `0' could be PAM_SILENT */
	  retcode = pam_open_session(pamh,0);
	  bail_out(0,retcode,"pam_open_session");
	  if (retcode != PAM_SUCCESS) {
	       fprintf(stderr,"%s: problem opening a session\n",argv[0]);
	       break;
	  }

	  fprintf(stderr,"The user has been authenticated and `logged in'\n");
	  fprintf(stderr,"This is where one would set\n"
		  "the credentials of the user...\n");

	  /* close a session for the user --- `0' could be PAM_SILENT
	   * it is possible that this pam_close_call is in another program..
	   */

	  retcode = pam_close_session(pamh,0);
	  bail_out(0,retcode,"pam_close_session");
	  if (retcode != PAM_SUCCESS) {
	       fprintf(stderr,"%s: problem closing a session\n",argv[0]);
	       break;
	  }
	  
	  break;                      /* don't go on for ever! */
     }

     /* close the Linux-PAM library */
     retcode = pam_end(pamh, PAM_SUCCESS);
     if (retcode)
	  bail_out(1,retcode,"pam_end");

     exit(0);
}
